.citation-list {
    font-size: .8rem;

    .label {
        font-weight: bold;
    }

    p {
        font-style: italic;
    }
}

dl {
    &.option-list,
    &.simple {
        kbd {
            border: unset;
            font-size: 1rem;
        }
    }
}

p code.literal {
    border: unset;
}

pre {
  word-wrap: break-word;
  white-space: pre-wrap;
  white-space: -moz-pre-wrap;
  white-space: -pre-wrap;
  white-space: -o-pre-wrap;
}

